# -*- coding: utf-8 -*-
"""
Created on Mon Aug 20 09:47:57 2018

@author: th0lxl1
"""

from numpy import sqrt, tile, transpose,  sin, exp, zeros
from numpy import linspace
from numpy import sum as nsum
from numpy import abs as nabs
import xraydb as xdb
import numpy as np
import scipy.constants as scc


def mlayer_rough(alpha,k0,n_r,zm0,sig):        
    """        
    mlayer(alpha,n,z,sig)
    Updated LBL May 30, 2022 
    from M. Tolan
    X-Ray Scattering from Soft-Matter Thin Films
    Chapter 2, p12 and following

    multi-layer parratt model with roughness
    
    We model the surface as a number of interfaces.  To  avoid confusion, 
    the electron densities are put in as electron density differences.  
    
    Variables:
    alpha --- The incident angle in radians
    k0    --- The magnitude of the wavevector of the light
    n     --- The  indices of refraction 
    z     --- The positions of the interfaces 
    sig   --- The roughness of each interface

    Translated to python August 20, 2018, Laurence Lurio 
    
    Note that R[:,0] is the top layer reflectivity and that T[:,0]
    is meaningless (just equal to 1) T[:,1] is the transmission from
    the first layer to the second layer
    
    O is the top interface
    
    Note index of refraction must be 1 - delta + i beta
    
    Note: T and R stand for the electric fields, not the
    coefficients.  Their form contains the z dependence, e.g
    R exp(+ikz)  and T exp(-ikz), with z a negative number going
    into the material
    """
    la = alpha.shape[0]
    ln = n_r.shape[0] # ln is the number of layers
    # need to subtract off the first index, as Tolan formalism assumes 
    # you are coming in from vacuum
    nr2d = n_r**2-n_r[0]**2
    nr2d = tile(nr2d,(la,1))
    nr2d = nr2d.astype(complex)
    zm = tile(zm0,(la,1))
    sig = tile(sig,(la,1))
    alpha = transpose(tile(alpha,(ln,1)))
    R = zeros((la,ln),dtype=complex)
    T = zeros((la,ln),dtype=complex)
    X = zeros((la,ln),dtype=complex)
    kz=k0*sqrt(sin(alpha)**2 + nr2d)
    rr=(kz[:,0:-1]-kz[:,1:])/(kz[:,0:-1]+kz[:,1:]) # eq. 2.17
    # Here is where the roughness factors in
    rr=rr*exp(-nabs(2*kz[:,0:-1]*kz[:,1:]*sig*sig))
    # tp=tp*exp(nabs((kz[:,0:-1]-kz[:,1:])**2*sig*sig/2))
    tp = 1-rr
    # append 0 and 1 for the reflection and transmission coefficient of the last layer
    
    for jl in range(ln-2,-1,-1):
        # since rr only describes the interfaces, not the layers,
        # the number of rr's is 1 less than the number of kz's
        X[:,jl]=rr[:,jl]+X[:,jl+1]*exp(2j*kz[:,jl+1]*zm[:,jl])
        X[:,jl]=X[:,jl]*exp(-2j*kz[:,jl]*zm[:,jl])
        X[:,jl]=X[:,jl]/(1+rr[:,jl]*X[:,jl+1]*exp(2j*kz[:,jl+1]*zm[:,jl]))
    T[:,0]=1.0
    R[:,0]=X[:,0]
    for jl in range(0,ln-1):
        R[:,jl+1]=(R[:,jl]*exp(-1j*(kz[:,jl+1]-kz[:,jl])*zm[:,jl])
                   -T[:,jl]*rr[:,jl]*exp(-1j*(kz[:,jl+1]+kz[:,jl])*zm[:,jl]))
        R[:,jl+1]=R[:,jl+1]/tp[:,jl]
        T[:,jl+1]=(T[:,jl]*exp(1j*(kz[:,jl+1]-kz[:,jl])*zm[:,jl])
                   -R[:,jl]*rr[:,jl]*exp(1j*(kz[:,jl+1]+kz[:,jl])*zm[:,jl]))
        T[:,jl+1]=T[:,jl+1]/tp[:,jl]
    return([T,R,kz,X,rr,tp,zm0])

def mlayer_conv(alpha,k0,n,z,sig,res,npt):
    dX = linspace(-res, res, num=npt)
    yout=[]
    mu=res/2.35
    norm = nsum(exp(-(dX/mu)**2))
    yout=0
    for delx in dX:
        _,R,_,_,_,_ = mlayer_rough(abs(alpha+delx),k0,n,z,sig)
        yout = yout+exp(-(delx/mu)**2)*abs(R[:,0])**2/norm
    return(yout)

def reflection_matrix(alpha,E,layers):
    '''
    reflection_matrix(alpha,E,layers)

    Parameters
    ----------
    alpha : numpy array float 
        incident angles (in radians).
    E : float
        incident energy (eV).
    layers : array of layers 
        each layer has a material (e.g. 'H2O') and density (g/cc) a thickness
        and a roughness.  
    The sig and thickness for the zeroth layer are thrown away as they have no meaning.

    Returns
    ----------
    T : numpy array (n_alpha, n_layer)
        Transmission matrix returned by mlayer or mlayer_rough
    R : numpy array (n_alpha, n_layer)
        Reflectivity matrix returned by mlayer or mlayer_rough.
    kz : numpy array (n_alpha, n_layer)
        wavevector matrix returned by mlayer or mlayer_rough. (units of inv meters)
    '''   
    k0 = 2*np.pi*E*scc.e/scc.h/scc.c
    nl  = len(layers)
    n_r = np.zeros(nl,complex)
    h_i = np.zeros(nl-1)
    sig_i = np.zeros(nl-1)
    z = 0
    for  i, (material, density, thick, rough)   in enumerate(layers):
        delta,beta,att = xdb.xray_delta_beta(material,density,E)
        n_r[i] = 1-delta+1j*beta 
        if i>0:
            h_i[i-1] = z
            z -= thick*scc.angstrom;
            sig_i[i-1] = rough*scc.angstrom
    T,R,kz,X,rr,tp,zd  = mlayer_rough(alpha,k0,n_r,h_i,sig_i)
    return T,R,kz,X,rr,tp,zd


def standing_wave(z,T,R,kz,h_i):
    '''
    standing_wave(T,R,kz,z)
    
    Function to calculate the electric field standing wave 
    
    Parameters
    ----------
    z : numpy array (1d)
        heights relative to top surface at which to calculate
            the standing wave.
    T : numpy array (n_alpha, n_layer)
        Transmission matrix returned by mlayer or mlayer_rough
    R : numpy array (n_alpha, n_layer)
        Reflectivity matrix returned by mlayer or mlayer_rough.
    kz : numpy array (n_alpha, n_layer)
        wavevector matrix returned by mlayer or mlayer_rough.
    h_i : numpy array (n_layer-1)
        array of interface height positions

            
    Here n_alpha is the number of angles input to mlayer and n_layer the
    number of layers input to mlayer.

    Returns
    -------
    I : numpy array (n_z,n_alpha)
        Standing wave intensity.
    E : numpy array (n_z,n_alpha)
        Electric field intensity.

    '''
    n_alpha,n_layer = np.shape(T)
    n_z = len(z)
    E = np.zeros((n_z,n_alpha))+0j   
    for i in range(n_layer):
        if i == 0:
            wz = z >= h_i[i]
        elif i == n_layer-1:
            wz = z < h_i[i-1]
        else :
            wz = (z < h_i[i-1])*(z >= h_i[i])
        # only do calculation for subset of z values
        tz = z[wz]
        n_tz = len(tz)
        T_full = np.broadcast_to(T[:,i],(n_tz,n_alpha))
        R_full = np.broadcast_to(R[:,i],(n_tz,n_alpha))
        kz_full = np.broadcast_to(kz[:,i],(n_tz,n_alpha))
        z_full = np.transpose(np.broadcast_to(tz,(n_alpha,n_tz)),(1,0))
        
        E[wz,:] = np.exp(-1j*kz_full*z_full)*T_full
        E[wz,:] += np.exp(1j*kz_full*z_full)*R_full
    # expand all the arrays to the same size for multiplying   
    I = np.abs(E)**2
    return I,E


    
    



# -*- coding: utf-8 -*-
"""
Created on Fri Jun 10 08:41:22 2022

@author: lluri
"""
import xraydb as xdb
import scipy.constants as scc
import numpy as np
def eden_to_rho(material,eden):
    # converts from electron density to mass density
    atoms = xdb.chemparse(material)
    A = 0
    Z = 0
    for atom_name in atoms.keys():
        tA = xdb.atomic_mass(atom_name)
        tZ = xdb.atomic_number(atom_name)
        A += tA*atoms[atom_name]
        Z += tZ
    rho = A*eden/scc.Avogadro/1e6/Z
    return rho
def rho_to_n(material,rho,energy):
    delta, beta, _ = xdb.xray_delta_beta(material, rho, energy)
    n = 1-delta - 1j*beta
    return n
def rho_to_rhoe(material,rho,energy):
    lam = scc.h*scc.c/scc.e/energy
    k = 2*np.pi/lam
    delta, beta, _ = xdb.xray_delta_beta(material, rho, energy)
    r0 = scc.physical_constants['classical electron radius'][0]
    rhoe = delta*k**2/2/np.pi/r0
    return rhoe



# Model for gold fluorescence
from lmfit import Model, models, create_params, Parameters
from matplotlib import pyplot as plt
import numpy as np
import scipy.constants as scc
import importlib.resources
from lmfit import Parameters
import xraydb as xdb
from pySWXF import fluor_fit, refl_funs
from pySWXF.fluor_fit import multilayer_ref_new_model, multilayer_model_Ti
from pySWXF import spec_utils, AuI_funs
import re
import pandas as pd
from scipy.interpolate import RectBivariateSpline
from typing import Dict, Any, List, Tuple
import sys

g1 = models.GaussianModel(prefix='g1')
g2 = models.GaussianModel(prefix='g2')
q1 = models.QuadraticModel() 
peak_model = g1+g2+q1

# code to fit reflectivity and find offset

MULTILAYER_FIT_NEW = 'NM_clean_NIU_NOKW_Mar26_24_fit4.json'
MULTILAYER_FIT_OLD = 'niu_multilayer_fit_params_3_LBL.json'

def water_refract(th,Beam_Energy):
        alphac_water = np.sqrt(2*xdb.xray_delta_beta('H2O',1.00,Beam_Energy)[0])
        thc_water = alphac_water/scc.degree
        return np.sqrt(th**2  - thc_water**2)


def get_params(fitflag):
    if fitflag == 'NN':
        parfilename = MULTILAYER_FIT_NEW
    elif fitflag == 'OO':
        parfilename = MULTILAYER_FIT_OLD
    elif fitflag == 'NO':
        parfilename = MULTILAYER_FIT_OLD
    else:
        print(f'get_params Not configured for flag {fitflag}')
        sys.exit()
    with importlib.resources.open_text('pySWXF', parfilename) as f:
        params = Parameters()
        params.load(f)
    return(params)


def get_offset_fit(th, I, dI, dinfo, scanno,  varylist = None, startlist = None, showfit=True, verbose=True):
    """
    Fit the offset in the reflection data for different experimental setups.

    Parameters:
    -----------
    th : array-like
        The angular positions of the measurements.
    I : array-like
        The measured reflected intensity.
    dI : array-like
        The uncertainties in the measured intensities (not used in the function, but included for completeness).
    showfit : bool, optional
        Whether to display the fit plot. Default is True.
    verbose : bool, optional
        Whether to print detailed information during execution. Default is True.

    Returns:
    --------
    thoff : float
        The angular offset resulting from the fit.
    result : lmfit.model.ModelResult
        The result of the fit, containing the best-fit parameters and statistics.
    """
    
    params = AuI_funs.get_params(dinfo['fitflag'])
    
    if dinfo['fitflag'] == 'NN':
        if verbose: 
            print(f'Using flag {dinfo["fitflag"]}: New cell, New multilayer')
        th_refrac = AuI_funs.water_refract(th, Beam_Energy)
        fitfun = AuI_funs.multilayer_ref_new_model
    elif dinfo['fitflag'] == 'OO':
        if verbose: 
            print(f'Using flag {dinfo["fitflag"]}: Old cell, Old multilayer')
            print('Side entry sample cell: No refraction correction.')
        th_refrac = th
        fitfun = AuI_funs.multilayer_model_Ti
    elif dinfo['fitflag'] == 'NO':
        if verbose: 
            print(f'Using flag {dinfo["fitflag"]}: New cell, Old multilayer')
            print('Top entry sample cell: Using refraction correction.')
        th_refrac = AuI_funs.water_refract(th, Beam_Energy)
        fitfun = AuI_funs.multilayer_model_Ti
    else:
        print(f'Not configured for flag {dinfo["fitflag"]}')
        sys.exit()
    
    for key in params.keys():
        params[key].vary = False

    params['I0'].vary = True
    params['I0'].max = 1e9
    params['I0'].min = 100
    params['I0'].value = np.max(I) * 3.5
    params['thoff'].vary = True
    params['thoff'].min = -0.1
    params['thoff'].max = 0.1
    params['res'].value = 0.001

    if varylist is not None:
        for item in varylist:
            params[item].vary = True
    if startlist is not None:
        for item in startlist:
            params[item[0]].value = item[1]

    if verbose: 
        print('Fitting amplitude and offset')

    presim = fitfun.eval(params=params, theta=th_refrac, 
                         Energy=dinfo['Energy'], water=True,
                         bilayer=True, new_cell=False)
    cen_sim = np.sum(presim * th) / np.sum(presim)
    cen_dat = np.sum(I * th) / np.sum(I)
    params['thoff'].value = cen_dat - cen_sim

    result = fitfun.fit(I, theta=th_refrac, params=params, 
                        Energy=dinfo['Energy'], water=True, bilayer=True,
                        new_cell=False, weights=I * 0 + 1)
    thoff = result.params['thoff'].value
    
    th_peak = th[np.argmax(I)]
    
    if showfit:    
        ysim = result.eval(theta=th_refrac, Energy=dinfo['Energy'],
                           water=True, bilayer=True, new_cell=False)
        plt.plot(th, I, label='data')
        plt.plot(th, ysim, '-k', label='fit')
        plt.locator_params(axis='x', nbins=20)
        plt.grid()
        plt.xlabel('th (deg)')
        plt.ylabel('Reflected Intensity')
        plt.title(f'{dinfo["spec_filename_stub"]:s} scan {scanno:d}')
        plt.legend()
        if verbose:
            print(f'Angular offset = {thoff:.3f}') 
    return thoff, th_peak, result

def setup_fit(dinfo,mca_sum):
    """
    Set up the fitting function and initial parameters based on the fluorophore.

    Parameters:
    - dinfo (dict): Dictionary containing experimental information.

    Returns:
    - fit_fun (Model): Combined lmfit model.
    - pars (Parameters): Initial parameters for the model.
    - data_range (array-like): Range of data to fit.
    - amplitude_names (list): List of amplitude parameter names.
    """
    if (dinfo['fluorophore'] == 'Br') and  (dinfo['run_date'] == 'Nov23'):
        return setup_fit_br_nov23(dinfo,mca_sum)
    elif dinfo['fluorophore'] == 'gold':
        return setup_fit_gold(dinfo,mca_sum)
    elif (dinfo['fluorophore'] == 'Br') and (dinfo['run_date'] == 'Feb24'):
        return setup_fit_br_feb24(dinfo,mca_sum)
    elif (dinfo['fluorophore'] == 'Br') and (dinfo['run_date'] == 'Mar23'):
        return setup_fit_br_mar23(dinfo,mca_sum)

def setup_fit_br_nov23(dinfo,mca_sum):
    quad_mod = models.QuadraticModel()
    gaus_mod = models.GaussianModel()
    fit_fun = quad_mod + gaus_mod
    pars = Parameters()
    pars.add('amplitude', value=150000, min=0)
    pars.add('center', value=13400)
    pars.add('sigma', value=115, vary=False)
    pars.add('a', value=0)
    pars.add('b', value=0)
    pars.add('c', value=5000)
    data_range = (dinfo['E'] > 13359 - 756.3) & (dinfo['E'] < 13359 + 302.6)
    amplitude_names = ["amplitude"]
    return fit_fun, pars, data_range, amplitude_names

def setup_fit_gold(dinfo,mca_sum):
    g1 = models.GaussianModel(prefix='g1_')
    g2 = models.GaussianModel(prefix='g2_')
    q1 = models.QuadraticModel()
    fit_fun = g1 + g2 + q1
    pars = Parameters()
    pars.add('g1_center', value=13400, vary=False)
    pars.add('g2_center', value=13771, vary=False)
    pars.add('g1_sigma', value=114.8, vary=False)
    pars.add('g2_sigma', value=114.8, vary=False)
    pars.add('g1_amplitude', value=2e5)
    pars.add('g2_amplitude', value=4e4)
    pars.add('a', value=0)
    pars.add('b', value=-7.5)
    pars.add('c', value=5000)
    data_range = (dinfo['E'] > 13000) & (dinfo['E'] < 14100)
    amplitude_names = ["g1_amplitude", "g2_amplitude"]
    return fit_fun, pars, data_range, amplitude_names

def setup_fit_br_feb24(dinfo,mca_sum):
    data_range = (dinfo['E'] > 11000) & (dinfo['E'] < 12500)
    mx = np.max(mca_sum[data_range])
    RR = mx/300
    g1 = models.GaussianModel(prefix='g1_')
    g2 = models.GaussianModel(prefix='g2_')
    q1 = models.QuadraticModel()
    fit_fun = g1 + g2 + q1
    pars = Parameters()
    pars.add('g1_center', value=11920, vary=True)
    pars.add('g2_center', value=11490, vary=True)
    pars.add('g1_sigma', value=114.8, vary=False)
    pars.add('g2_sigma', value=114.8, vary=False)
    pars.add('g1_amplitude', value=3876*RR)
    pars.add('g2_amplitude', value=51770*RR)
    pars.add('a', value=8.897e-06*RR)
    pars.add('b', value=-0.211*RR)
    pars.add('c', value=1268*RR)
    amplitude_names = ["g1_amplitude"]
    return fit_fun, pars, data_range, amplitude_names

def setup_fit_br_mar23(dinfo,mca_sum):
    data_range = (dinfo['E'] > 11000) & (dinfo['E'] < 12500)
    mx = np.max(mca_sum[data_range])
    RR = mx/300
    g1 = models.GaussianModel(prefix='g1_')
    g2 = models.GaussianModel(prefix='g2_')
    q1 = models.QuadraticModel()
    fit_fun = g1 + g2 + q1
    pars = Parameters()
    pars.add('g1_center', value=11920, vary=True)
    pars.add('g2_center', value=11490, vary=True)
    pars.add('g1_sigma', value=114.8, vary=False)
    pars.add('g2_sigma', value=114.8, vary=False)
    pars.add('g1_amplitude', value=3876*RR)
    pars.add('g2_amplitude', value=51770*RR)
    pars.add('a', value=8.897e-06*RR)
    pars.add('b', value=-0.211*RR)
    pars.add('c', value=1268*RR)
    amplitude_names = ["g1_amplitude"]
    return fit_fun, pars, data_range, amplitude_names

def get_fluor_amplitude(mca_instance, dinfo, start_pars=None):
    """
    Get the fluorescence amplitude from the data.

    Parameters:
    - mca_data (array-like): MCA data to fit.
    - dinfo (dict): Dictionary containing experimental information.
    - start_pars (Parameters, optional): Starting parameters for the fit.

    Returns:
    - peak_counts (float): Sum of peak counts.
    - peak_errs (float): Sum of peak errors.
    """
    fit_fun, pars, data_range, amplitude_names = setup_fit(dinfo,mca_instance)
    if start_pars is not None:
        pars = start_pars
        for amplitude_name in amplitude_names:
            pars[amplitude_name].max = pars[amplitude_name].value * 4
            pars[amplitude_name].min = 0
    dE = dinfo['E'][1] - dinfo['E'][0]
    X = dinfo['E'][data_range]
    this_result = fit_fun.fit(mca_instance[data_range] / dE, x=X, params=pars)
    peak_counts = sum(this_result.params[amplitude_name].value for amplitude_name in amplitude_names)
    peak_errs = np.sqrt(sum(
        (this_result.params[amplitude_name].stderr if isinstance(this_result.params[amplitude_name].stderr, float) else np.inf)**2
        for amplitude_name in amplitude_names
    ))
    return peak_counts, peak_errs

def get_start_pars(mca_data, dinfo):
    """
    Get the starting parameters for the fit.

    Parameters:
    - mca_data (array-like): MCA data to fit.
    - dinfo (dict): Dictionary containing experimental information.

    Returns:
    - pars (Parameters): Fitted parameters.
    - this_result (ModelResult): Result of the initial fit.
    - amplitude_names (list): List of amplitude parameter names.
    """
    mca_sum = np.sum(mca_data, 0)
    fit_fun, pars, data_range, amplitude_names = setup_fit(dinfo, mca_sum)
    dE = dinfo['E'][1] - dinfo['E'][0]
    X = dinfo['E'][data_range] 
    npt = mca_data.shape[0]
    this_result = fit_fun.fit(mca_sum[data_range] / dE, x=X, params=pars)
    pars = this_result.params
    for amplitude_name in amplitude_names:
        pars[amplitude_name].value /= npt
    return pars, this_result, amplitude_names

def show_fluor_fit(result, dinfo, amplitude_names, label=True):
    """
    Show the fit result for fluorescence data.

    Parameters:
    - result (ModelResult): Result of the fit.
    - dinfo (dict): Dictionary containing experimental information.
    """
    result.fit()
    result.plot_fit()
    E = result.userkws['x']
    ysim0 = result.eval()
    par0 = result.params
    for amplitude_name in amplitude_names:
        par0[amplitude_name].value = 0
    ysim1 = result.eval(params=par0)
    plt.plot(E, ysim0 - ysim1, '--g', label='Fluorescence Intensity')
    plt.legend()
    plt.ylabel('Counts/Second')
    plt.xlabel('Energy (eV)')
    plt.title(f'Fit to Fluorescence Peak {dinfo["fluorophore"]}')
    if label:
        spec_utils.K_label('Br', height=.8)
        spec_utils.L_label('Au', height=.5)


def get_fluorescence_data(mca_data, dinfo, norm=None, plot_fit=False):
    """
    Get fluorescence data and fit results.

    Parameters:
    - mca_data (array-like): MCA data to fit.
    - dinfo (dict): Dictionary containing experimental information.
    - norm (array-like, optional): Normalization factor.
    - plot_fit (bool, optional): Whether to plot the fit.

    Returns:
    - mca_amplitudes (array-like): MCA amplitudes.
    - mca_errors (array-like): MCA errors.
    """
    nscanpoints = mca_data.shape[0]
    mca_amplitudes = np.empty(nscanpoints)
    mca_errors = np.empty(nscanpoints)
    start_pars, start_result, amplitude_names = get_start_pars(mca_data, dinfo)
    if plot_fit:
        show_fluor_fit(start_result, dinfo, amplitude_names)
    for ii in range(nscanpoints):
        A, dA = get_fluor_amplitude(mca_data[ii, :], dinfo, start_pars=start_pars)
        mca_amplitudes[ii] = A
        mca_errors[ii] = dA
    if norm is not None:
        norm_factor = norm / np.average(norm)
        mca_amplitudes *= norm_factor
        mca_errors *= norm_factor
    return mca_amplitudes, mca_errors, start_pars

def plot_fluorescence(angles, mca_amplitudes, mca_errors, dinfo, fluorescence_scan_number, nflist=None):
    """
    Plot fluorescence data.

    Parameters:
    - angles (array-like): Angles of the measurement.
    - mca_amplitudes (array-like): MCA amplitudes.
    - mca_errors (array-like): MCA errors.
    - dinfo (dict): Dictionary containing experimental information.
    - fluorescence_scan_number (int): The scan number for fluorescence.
    - nflist (str, optional): List of scan numbers.
    """
    plt.errorbar(angles, mca_amplitudes, mca_errors, fmt='-ks')
    plt.xlabel('theta (deg)')
    plt.ylabel('fluorescence intensity (cps)')
    scanstring = nflist if nflist is not None else f'{fluorescence_scan_number}'
    plt.title(f"file: {dinfo['vortex_filename_stub']} scan: {scanstring}, {dinfo['fluorophore']}")
    plt.ylim(0.7 * np.min(mca_amplitudes), 1.3 * np.max(mca_amplitudes))

def fluor_multifit(N_align, N_fluor, dinfo):
    """
    Perform multi-fit for fluorescence data.

    Parameters:
    - N_align (list): List of alignment scan numbers.
    - N_fluor (list): List of fluorescence scan numbers.
    - dinfo (dict): Dictionary containing experimental information.

    Returns:
    - angles_0 (array-like): Final angles.
    - mca_amplitudes_0 (array-like): Final MCA amplitudes.
    - mca_errors_0 (array-like): Final MCA errors.
    - nflist (str): List of fluorescence scan numbers.
    - thoff (float): offset angle from first fit.
    """
    nflist = ", ".join(map(str, N_fluor))
    last_align_scan_number = 0
    for i, (align_scan_number, fluorescence_scan_number) in enumerate(zip(N_align, N_fluor)):
        print(f'Working on dataset {i+1} align scan {align_scan_number} fluor scan {fluorescence_scan_number}')
        
        if align_scan_number != last_align_scan_number:
            print('Fitting alignment scan for angle offset')
            # Get reflectivity data
            data, _ = spec_utils.readscan(dinfo["align_filename"], align_scan_number)
            th = np.array(data['MU'])
            I = np.array(data['L_ROI1'])
            dI = np.sqrt(I)
            rr = (th > 0.4) & (th < 0.6)
            theta_offset, _ = AuI_funs.get_offset_fit(th[rr], I[rr], dI[rr], dinfo["Energy_Nov"], dinfo["fitflag"], showfit=False, verbose=True)
        else:
            print('Repeated alignment scan. No fit required')
        
        last_align_scan_number = align_scan_number

        mca_data, mca_sum, ndset, angles, npt, exposure_time, norm = spec_utils.get_fluor_data(
            dinfo["spec_filename"], dinfo["mca_filename"], fluorescence_scan_number)
        
        mca_amplitudes, mca_errors = get_fluorescence_data(mca_data / exposure_time, dinfo, norm=None, plot_fit=False)
        
        if i == 0:
            thoff = theta_offset
            mca_amplitudes_0 = mca_amplitudes
            mca_errors_0 = mca_errors
            angles_0 = angles
        else:
            adjusted_angles = angles + theta_offset - thoff
            interpolated_amplitudes = np.interp(angles, adjusted_angles, mca_amplitudes)
            interpolated_errors = np.interp(angles, adjusted_angles, mca_errors)
            mca_amplitudes_0, mca_errors_0 = spec_utils.cbwe_s(mca_amplitudes_0, mca_errors_0, interpolated_amplitudes, interpolated_errors)

        nnorm = norm / np.mean(norm)
    return angles_0, mca_amplitudes_0 / nnorm, mca_errors_0 / nnorm, nflist, thoff



def get_Zlist(N,D):
    # D = bilayer thickness
    # N slabs in bilayer
    edgelist = np.linspace(0,D,N+1)        # positions of interfaces of slabs
    Zlist = (edgelist[0:-1]+edgelist[1:])/2   # positions of centers of slabs
    return Zlist, edgelist

def multilayer_fluor_lay_N(theta,Avec,Imap,zmax):
    ''' multilayer_fluor_lay_N(theta,I0,thoff,bg,Avec)
    breaks up bilayer into N slabs wit N the dimension of Avec
    The A's are the amplitudes of the slabs
    '''
    # need to add feature to convolute with angular resolution
    alpha = theta*scc.degree
    Zlist, edgelist = get_Zlist(np.size(Avec), zmax)
    Ifield = Imap(Zlist, alpha)
    # sum up the product of the fluoresence from each slab times the amplitude in the slab
    y = np.sum(Ifield*np.expand_dims(Avec,1),0)
    return(y)


def pars_to_layer(pars: Dict[str, Any], dinfo: Dict[str, Any]) -> Any:
    """
    Convert parameter values to layer information based on fit flag.

    Parameters:
    -----------
    pars : Dict[str, Any]
        Dictionary containing parameter values.
    dinfo : Dict[str, Any]
        Dictionary containing fitting information including the fit flag.

    Returns:
    --------
    layers : Any
        Layer information generated by the sample simulation function.
    """
    if dinfo['fitflag'] == 'NN':
        expected_params = {'d_overlay', 'sig_overlay', 'G', 'D', 'sig_W', 'sig_SiC', 'D_Cr', 'sig_Cr', 'sig_Si_Bot'}
        sample_fun = fluor_fit.sample_sim_new
        water_flag = True 
        bilayer_flag = True 
        new_cell_flag = False
    elif (dinfo['fitflag'] == 'OO')  or (dinfo['fitflag'] == 'NO'):
        expected_params = {'G', 'D', 'sig_Mo', 'sig_Si', 'overlay', 'sig_Si_top', 'D_Ti', 'sig_Ti', 'D_SiO2', 'sig_SiO2',
                           'D_SiO2_Bot', 'sig_SiO2_Bot', 'sig_Si_Bot'}
        sample_fun = fluor_fit.sample_sim
        water_flag = True 
        bilayer_flag = True 
        new_cell_flag = None
    else:
        raise ValueError(f"Unknown fitflag: {dinfo['fitflag']}")

    filtered_par = {key: value.value for key, value in pars.items() if key in expected_params}

    layers = sample_fun(**filtered_par, water=water_flag, bilayer=bilayer_flag, new_cell=new_cell_flag)
    return layers

def plot_N_slab_result(result,NUM_SLABS, zmax):
    """
    Plot the fluorophore concentration across three slabs up to a maximum height.

    Parameters:
    result : object containing simulation parameters and results
    zmax : float, the maximum height to consider for plotting
    """
    # Constants
    ANGSTROM = scc.angstrom  # This assumes scc has been properly imported

    # Unpacking parameters
    A = [result.params[f'A{i}'].value for i in range(NUM_SLABS)]
    dA = [result.params[f'A{i}'].stderr for i in range(NUM_SLABS)]
    _, edgelist = get_Zlist(NUM_SLABS, zmax)

    # Check that edgelist is sufficiently long
    if len(edgelist) < NUM_SLABS + 1:
        raise ValueError("edgelist does not contain enough entries.")

    # Plotting
    for i, (tA,tdA) in enumerate(zip(A,dA)):
        edge1 = edgelist[i] / ANGSTROM
        edge2 = edgelist[i + 1] / ANGSTROM
        CEN = (edge1+edge2)/2
        plt.plot([edge1, edge1], [0, tA], '-k')
        plt.plot([edge1, edge2], [tA, tA], '-k')
        plt.plot([edge2, edge2], [tA, 0], '-k')
        if tdA  is not None:
            plt.errorbar([CEN],[tA],[tdA],fmt='ks')

    # now plot error bars

    plt.xlabel('height ($\\mathrm{\\AA}$)')
    plt.ylabel('fluorophore concentration')
    plt.title('Fluorophore Concentration Profile')

def get_bilayer_position(layers: List[Tuple[str, float, float, float]]) -> Tuple[float, float]:
    """
    Calculate the position and thickness of the bilayer in the given list of layers.

    Parameters:
    -----------
    layers : List[Tuple[str, float, float, float]]
        A list of tuples, each containing (material, density, thickness, roughness).

    Returns:
    --------
    bilayer_top : float
        The depth at which the top of the bilayer is found.
    bilayer_thickness : float
        The total thickness of the bilayer.
    """
    depth = 0.0
    found_top = False
    bilayer_thickness = 0.0
    bilayer_top = 0.0

    for material, density, thickness, roughness in layers:
        if not found_top and material == 'CH2':
            bilayer_top = depth
            found_top = True

        depth -= thickness

        if material == 'CH2':
            bilayer_thickness += thickness

    return bilayer_top, bilayer_thickness

def get_standing_wave(
    dinfo: Dict[str, Any], 
    layers: Any, 
    bilayer_top: float, 
    bilayer_thickness: float, 
    result_reflectivity: Any, 
    thoff: float, 
    th_com: float, 
    plot_wave: bool = True
) -> RectBivariateSpline:
    """
    Calculate the standing wave field and optionally plot it.

    Parameters:
    -----------
    dinfo : Dict[str, Any]
        Dictionary containing fitting information and parameters.
    layers : Any
        Layer structure used in the reflection matrix calculation.
    bilayer_top : float
        The top position of the bilayer.
    bilayer_thickness : float
        The thickness of the bilayer.
    result_reflectivity : Any
        Result object containing fitted parameters.
    thoff : float
        The offset angle obtained from the reflectivity fit
    th_com : float
        The reflection center angle.
    plot_wave : bool, optional
        Whether to plot the standing wave field. Default is True.

    Returns:
    --------
    SW : RectBivariateSpline
        Bivariate spline object representing the standing wave field.
    """
    try:
        alpha = np.linspace(0.1, 1, 1000) * scc.degree

        if (dinfo['fitflag'] == 'NN') or (dinfo['fitflag'] == 'NO'):
            alphap = AuI_funs.water_refract(alpha / scc.degree, dinfo['Energy']) * scc.degree
        elif dinfo['fitflag'] == 'OO':
            alphap = alpha
        else:
            print('error: unknown fit flag')

        t, r, kz, X, rr, tp, zm = refl_funs.reflection_matrix(alphap, dinfo['Energy'], layers)
        heights0 = np.linspace(-1000, 500, 1000) * scc.angstrom

        print(f'using thoff = {thoff:f}')
        print(f'using energy = {dinfo["Energy"]:f}')
        print(f'using fitflag {dinfo["fitflag"]:s}')
        result_reflectivity.params.pretty_print()

        If, Ef = refl_funs.standing_wave(heights0, t, r, kz, zm)
        offset = (bilayer_top - bilayer_thickness) * scc.angstrom
        SW = RectBivariateSpline(heights0 - offset, alpha+thoff*scc.degree, If)

        alpha_sim = np.linspace(0.1, 0.8, 1000) * scc.degree
        heights_sim = np.linspace(-500, 500, 1000) * scc.angstrom
        Isim = SW(heights_sim, alpha_sim)

        if plot_wave:
            plt.imshow(
                Isim, aspect='auto', extent=(alpha_sim[0]/scc.degree, alpha_sim[-1]/scc.degree,
                                             heights_sim[0]/scc.angstrom, heights_sim[-1]/scc.angstrom),
                origin='lower'
            )
            plt.xlabel('MU (deg)')
            plt.ylabel('z (angstrom)')
            plt.title('Standing Wave Intensity')
            plt.plot([alpha_sim[0]/scc.degree, alpha_sim[-1]/scc.degree], [0, 0], '--r', label='Bilayer bottom')
            plt.plot([alpha_sim[0]/scc.degree, alpha_sim[-1]/scc.degree], [bilayer_thickness, bilayer_thickness], '-r', label='Bilayer top')

            d_overlay = result_reflectivity.params.get('d_overlay', result_reflectivity.params.get('overlay')).value
            plt.plot([alpha_sim[0]/scc.degree, alpha_sim[-1]/scc.degree], [-d_overlay, -d_overlay], '-y', label='Multilayer top')

            plt.plot([th_com, th_com], [heights_sim[0]/scc.angstrom, heights_sim[-1]/scc.angstrom], '-m', label='Reflection center')

            plt.legend()

        return SW
    except Exception as e:
        print(f"An error occurred: {e}")
        raise

# Model for three slabs

def three_slab(theta,A0,A1,A2,bg, Imap,zmax):
    return multilayer_fluor_lay_N(theta,[A0,A1,A2],Imap,zmax) + bg

three_slab_model = Model(three_slab, independent_vars = ['theta', 'Imap', 'zmax'])

def fit_three_layer(angles, mca_amplitudes, mca_errors, bilayer_thickness,  SW, nflist, dinfo, plot=True): 
    # Here we fit the measured fluorescence to the calculated fluorescence
    angles = np.array(angles) 
    astart = np.mean(mca_amplitudes)/3.0
    params = three_slab_model.make_params(A0 =astart, A1 = astart, A2 = astart, bg = astart)
    params['A0'].vary=True
    params['A0'].min = 0
    params['A1'].vary=True
    params['A1'].min = 0
    params['A2'].vary=True
    params['A2'].min = 0
    params['bg'].vary = True 
    params['bg'].min = 0
    params['bg'].max = astart*1.5

    zmax = bilayer_thickness*scc.angstrom
    result3 = three_slab_model.fit(mca_amplitudes ,params = params,
                theta = angles, Imap = SW, zmax = zmax,
                weights = 1/mca_errors)
    thsim = np.linspace(angles[0],angles[-1],1000)
    ysim = result3.eval(theta = thsim, Imap = SW, zmax = zmax)
    if plot:
        plt.plot(thsim,ysim,'-r')
        plt.errorbar(angles,mca_amplitudes,mca_errors,fmt='-ks')
        print(result3.fit_report())
        plt.xlabel('angle (deg)')
        plt.ylabel('Fluorescence Intensity (counts)')
        plt.title(f'fluorescence scans: {nflist:s} file:{dinfo["vortex_filename_stub"]:s}')
        plt.xlim(.40,.52)
    return result3, zmax

# Model for five slabs
def five_slab(theta,A0,A1,A2,A3, A4, bg, Imap,zmax):
    return multilayer_fluor_lay_N(theta,[A0,A1,A2,A3,A4],Imap,zmax) + bg

five_slab_model = Model(five_slab, independent_vars = ['theta', 'Imap', 'zmax'])

def fit_five_layer(angles, mca_amplitudes, mca_errors, bilayer_thickness,  SW, nflist, dinfo, plot=True): 
    # Here we fit the measured fluorescence to the calculated fluorescence
    astart = np.mean(mca_amplitudes)/5.0
    angles = np.array(angles)
    params = five_slab_model.make_params(A0 =astart, A1 = astart, A2 = astart, A3 = astart, A4 = astart, bg = astart)
    params['A0'].vary=True
    params['A0'].min = 0
    params['A0'].max = 10*astart
    params['A1'].vary=True
    params['A1'].min = 0
    params['A1'].max = 10*astart
    params['A2'].vary=True
    params['A2'].min = 0
    params['A2'].max = 10*astart
    params['A3'].vary=True
    params['A3'].min = 0
    params['A3'].max = 10*astart
    params['A4'].vary=True
    params['A4'].min = 0
    params['A4'].max = 10*astart
    params['bg'].vary = True 
    params['bg'].min = 0
    params['bg'].max = 10*astart

    zmax = bilayer_thickness*scc.angstrom
    result5 = five_slab_model.fit(mca_amplitudes ,params = params,
                theta = angles, Imap = SW, zmax = zmax,
                weights = 1/mca_errors, method='differential_evolution')
    thsim = np.linspace(angles[0],angles[-1],1000)
    ysim = result5.eval(theta = thsim, Imap = SW, zmax = zmax)
    if plot:
        plt.plot(thsim,ysim,'-r')
        plt.errorbar(angles,mca_amplitudes,mca_errors,fmt='-ks')
        print(result5.fit_report())
        plt.xlabel('angle (deg)')
        plt.ylabel('Fluorescence Intensity (counts)')
        plt.title(f'fluorescence scans: {nflist:s} file:{dinfo["vortex_filename_stub"]:s}')
        plt.xlim(.40,.52)
    return result5, zmax

def kapton_correct(th: float, Iin: float, dinfo: Dict[str, Any]) -> float:
    """
    Corrects the input intensity (Iin) for kapton absorption.

    Parameters:
    th (float): The angle in degrees.
    Iin (float): The input intensity to be corrected.
    dinfo (dict): Dictionary containing relevant parameters:
        - 'fitflag' (str): Flag indicating fit status. If 'OO', no correction is applied.
        - 'Energy' (float): Energy value for material_mu calculation.
        - 'sample_L' (float): Length of the sample.
        - 'd_kapton' (float): Thickness of the kapton.

    Returns:
    float: Corrected intensity if 'fitflag' is not 'OO'. Otherwise, returns the input intensity (Iin).
    """
    if dinfo['fitflag'] == 'OO':
        return Iin

    kapton = xdb.find_material('kapton')
    mu = xdb.material_mu(kapton.formula, dinfo['Energy'], kapton.density, kind='total') / scc.centi
    alpha = th * scc.degree

    npt = 3  # Using 3 points for numerical integration
    x = np.linspace(0, dinfo['sample_L'], npt)
    trans = 0

    for tx in x:
        l1 = np.min(np.vstack((dinfo['d_kapton'] / np.sin(th * scc.degree), tx / np.cos(th * scc.degree))), axis=0)
        l2 = np.min(np.vstack((dinfo['d_kapton'] / np.sin(th * scc.degree), (dinfo['sample_L'] - tx) / np.cos(th * scc.degree))), axis=0)
        trans += np.exp(-mu * (l1 + l2))
    trans /= npt
    return Iin / trans

def plot_spec_scan(th, I, dinfo, scan_number):
    plt.plot(th,I, '-ks')
    plt.xlabel('th (deg)')
    plt.ylabel('I (count/monitor)')
    plt.title(f'{dinfo["spec_filename_stub"]:s} scan {scan_number:d}')

def plot_ref_and_fluor(th, mca_amplitudes, mca_errors, I, dinfo, fluorescence_scan_number):
    fig, ax1 = plt.subplots()
    color = 'tab:red'
    ax1.set_xlabel('$\\theta$ (deg)')
    ax1.set_ylabel('fluorecence counts', color=color)
    ax1.errorbar(th, mca_amplitudes, mca_errors, fmt = '-ks',  color=color)
    ax1.tick_params(axis='y', labelcolor=color)

    ax2 = ax1.twinx()  
    color = 'black'
    ax2.set_ylabel('Reflectivity (cps)', color=color)
    ax2.plot(th, I, color=color)
    ax2.tick_params(axis='y', labelcolor=color)
    tit =  f' {dinfo["vortex_filename_stub"]:s} scan {fluorescence_scan_number:d}'
    plt.title(tit)
    fig.tight_layout()  
    plt.show()

def no_offset_fluor_merge(scanlist, dinfo):
    nflist = ", ".join(map(str, scanlist))
    for ii, scan in enumerate(scanlist):
        mca_data, mca_sum, ndset, angles, npt, exposure_time, norm = spec_utils.get_fluor_data(dinfo, scan)
        mca_amplitudes, mca_errors, start_pars = AuI_funs.get_fluorescence_data(mca_data/exposure_time, dinfo, norm = None, plot_fit = False)
        if ii==0:
            M_Y = mca_amplitudes
            DM_Y = mca_errors**2
        else:
            M_Y += mca_amplitudes
            DM_Y += mca_errors**2

    DM_Y = np.sqrt(DM_Y)**angles/.47
    M_Y = M_Y*angles/.47    
    return M_Y, DM_Y, nflist


def fluor_multifit_Br(N_align, N_fluor, dinfo):
    """
    Perform multifit for fluorescence data.

    Parameters:
    - N_align: List of alignment scan numbers.
    - N_fluor: List of fluorescence scan numbers.
    - dinfo: Dictionary containing relevant data information.

    Returns:
    - angles: Array of angles.
    - mca_amplitudes: Array of MCA amplitudes.
    - mca_errors: Array of MCA errors.
    - nflist: Comma-separated string of N_fluor values.
    """
    nflist = ", ".join(map(str, N_fluor))
    
    # Loop through all the data, offset and sum
    for i, (align_scan_number, fluorescence_scan_number) in enumerate(zip(N_align, N_fluor)):
        print(f'Working on dataset {i+1} align scan {align_scan_number} fluor scan {fluorescence_scan_number}')
        
        # Get reflectivity data
        data, _ = spec_utils.readscan(dinfo["align_filename"], align_scan_number)
        th = np.array(data['MU'])
        I = np.array(data['L_ROI1'])
        dI = np.sqrt(I)
        
        rr = (th>.4)*(th<.6)
        theta_offset, _ = AuI_funs.get_offset_fit(th[rr], I[rr], dI[rr], dinfo["Energy_Nov"], dinfo["fitflag"], showfit=False, verbose=True)
        
        # Get fluorescence data
        mca, _, _, angles, _, _, norm = spec_utils.get_fluor_data(dinfo["spec_filename"], dinfo["mca_filename"], fluorescence_scan_number)
        
        nscanpoints = mca.shape[0]
        mca_amplitudes = np.empty(nscanpoints)
        mca_errors = np.empty(nscanpoints)
        
        for ii in range(nscanpoints):
            A, dA = AuI_funs.get_gold_amplitude(dinfo["E"], mca[ii, :])
            mca_amplitudes[ii] = A
            mca_errors[ii] = dA
        
        norm_factor = norm / np.average(norm)
        mca_amplitudes *= norm_factor
        mca_errors *= norm_factor
        
        if i == 0:
            thoff = theta_offset
            mca_amplitudes_0 = mca_amplitudes
            mca_errors_0 = mca_errors
            angles_0 = angles
        else:
            adjusted_angles = angles + theta_offset - thoff
            interpolated_amplitudes = np.interp(angles, adjusted_angles, mca_amplitudes)
            interpolated_errors = np.interp(angles, adjusted_angles, mca_errors)
            mca_amplitudes_0, mca_errors_0 = spec_utils.cbwe_s(mca_amplitudes_0, mca_errors_0, interpolated_amplitudes, interpolated_errors)
    
    return angles_0, mca_amplitudes_0, mca_errors_0, nflist



import pandas as pd
import re
import numpy as np
import pandas as pd
from scipy.optimize import fsolve
import scipy.constants as scc
from lmfit.models import LinearModel, GaussianModel
import xraydb as xdb
from numpy import zeros
from lmfit import Model
import os
from matplotlib import pyplot as plt
from matplotlib.patches import FancyArrowPatch
def readscan(specfile,scanno):
    '''
    readscan(specfile,scanno)
    Input: 
    specfile string
    scanno integer

    Returns: 
     data: Array of column information in scans
     scan_info: Dictionary of columns, title, mvals
     columns are the column names
     title is the scan title
     mvals are the values of all the moters

    '''
    with open(specfile,'r') as fp:
            mpairs = []
            break_out = False
            for i, line in enumerate(fp):
                if break_out:
                    break
                if '#O' in line:
                    tmnames = line[4:-1].split('  ')
                    for ttmnames in tmnames:
                        if len(ttmnames) > 0:
                            mpairs.append([ttmnames,0.0])
                if '#S {0:d}'.format(scanno) in line:
                    title = line
                    nlines = int(line.split()[-2])+1
                    nother = 0
                    mindex = 0
                    nrows = 0 
                    filestart = i+1
                    start = False
                    for line  in fp:
                        if break_out:
                            break
                        if not start:
                            if '#L' in line:
                                cnames = re.split('\s{2,}',line[3:-1])   
                                nother += 1
                            elif '#P' in line:
                                tmvals = line[4:-1].split(' ')
                                for ttmval in tmvals:
                                    if len(ttmval) > 0:
                                        mpairs[mindex][1] += float(ttmval)
                                        mindex += 1
                                nother +=1
                            elif ('#' in line):
                                nother += 1
                            elif (not '#' in line):
                                nrows += 1
                                start = True
                        else:
                            if ('#C' in line):
                                break_out = True
                                break
                            elif (nrows == nlines):
                                break_out = True
                                break
                            else:
                                nrows += 1
    # clean up mvals
    mvals = {}
    for tpair in mpairs:
        mvals[tpair[0]] = tpair[1]
    scan_info = {'columns':cnames[1:],'title':title,'mvals':mvals}
    if nrows == 0:
        print('scan aborted')
        return([],scan_info)
    else:
        data = pd.read_csv(specfile,skiprows=filestart+nother,
                       nrows=nrows,delim_whitespace=True,header=None)                                
        data.columns=cnames        
        return(data,scan_info)
    
def readmcascan(specfile,scanno):
    with open(specfile,'r') as fp:
        # read down to start of scan
        start = False
        mca = False
        mca_i = 0
        nmca = 0
        mca_data = []
        M_mca_data = []
        start_chan = 0
        end_chan = 0
        data = []
        for i, line in enumerate(fp):
            if not start and '#S {0:d} '.format(scanno) in line:
                print(f'found scan {scanno:d} at line {i:d}')
                start = True
                continue
            if start and '#@CHANN ' in line:
                channels = re.split('\s{1,}',line[8:-1])
                start_chan = int(channels[1])
                end_chan = int(channels[2])
                print(f'start chan {start_chan:d} end_chan {end_chan:d}')
                continue
            if start and '#@MCA ' in line:
                data_per_line = int(line[-4:-2])
                continue
            if start and '@A' in line:
                mca = True
                nmca = (end_chan-start_chan)+1
                mca_data = np.zeros(nmca)
                mca_i = 0
            if start and mca:
                if mca_i == 0:
                    ifst = 3
                else:
                    ifst = 1
                if ord(line[-1]) == 10:
                    line = line[:-1]
                if ord(line[-1]) == 92:
                    line = line[:-1]
                tdata = list(map(int,line[ifst:].split())) 
                n1 = mca_i+data_per_line
                n2 = nmca 
                mca_data[mca_i:min(n1,n2)] = tdata
                mca_i += np.size(tdata)
                if mca_i == nmca:
                    M_mca_data.append(mca_data)
                    mca = False
                continue
            if start and not mca and not '#' in line:
                tdata = list(map(float,line[0:-1].split()))
                data.append(tdata)
            if start  and '#S' in line:
                start = False
    data['end channel'] = end_chan
    data['start_chan'] = start_chan
    return(M_mca_data,data) 

def readscan_2_original(specfile,scanno):
    with open(specfile,'r') as fp:
        # read down to start of scan
        start = False
        mca = False
        mca_i = 0
        nmca = 0
        mca_data = []
        M_mca_data = []
        start_chan = 0
        end_chan = 0
        data = []
        for i, line in enumerate(fp):
            if not start and '#S {0:d} '.format(scanno) in line:
                print(f'found scan {scanno:d} at line {i:d}')
                start = True
                title = line
                continue
            if start and '#L' in line:
                cnames = re.split('\s{2,}',line[3:-1]) 
                continue
            if start and '#@CHANN ' in line:
                channels = re.split('\s{1,}',line[8:-1])
                start_chan = int(channels[1])
                end_chan = int(channels[2])
                continue
            if start and '@A' in line:
                mca = True
                nmca = (end_chan-start_chan+1)
                mca_data = np.zeros(nmca)
                mca_i = 0
            if start and mca:
                if mca_i == 0:
                    ifst = 3
                else:
                    ifst = 1
                if ord(line[-1]) == 10:
                    line = line[:-1]
                if ord(line[-1]) == 92:
                    line = line[:-1]
                tdata = list(map(int,line[ifst:].split()))   
                mca_data[mca_i:min(mca_i+16,nmca)] = tdata
                mca_i += np.size(tdata)
                if mca_i == nmca:
                    M_mca_data.append(mca_data)
                    mca = False
                continue
            if start and not mca and not '#' in line:
                tdata = list(map(float,line[0:-1].split()))
                if len(tdata) == len(cnames):
                    data.append(tdata)
            if start  and '#S' in line:
                start = False
        pdata = pd.DataFrame(data)
        pdata.columns = cnames
        scan_info = {'columns':cnames[1:],'title':title}
    return(pdata,scan_info,M_mca_data,start_chan,end_chan)

def readscan_2(specfile, scanno):
    """
    Reads a scan from a spec file.

    Parameters:
    - specfile (str): The path to the spec file.
    - scanno (int): The scan number to read.

    Returns:
    - pdata (pd.DataFrame): Data from the scan as a pandas DataFrame.
    - scan_info (dict): Information about the scan, including column names and title.
    - M_mca_data (list): List of MCA data arrays.
    - start_chan (int): Starting channel.
    - end_chan (int): Ending channel.
    """
    try:
        with open(specfile, 'r') as fp:
            # Initialize variables
            start = False
            mca = False
            mca_i = 0
            nmca = 0
            mca_data = []
            M_mca_data = []
            start_chan = 0
            end_chan = 0
            data = []
            
            for i, line in enumerate(fp):
                line = line.rstrip()
                
                if not start and f'#S {scanno} ' in line:
                    print(f'Found scan {scanno} at line {i}')
                    start = True
                    title = line
                    continue
                
                if start and line.startswith('#L'):
                    cnames = re.split(r'\s{2,}', line[3:])
                    continue
                
                if start and line.startswith('#@CHANN '):
                    channels = re.split(r'\s+', line[8:])
                    start_chan = int(channels[1])
                    end_chan = int(channels[2])
                    continue
                
                if start and line.startswith('@A'):
                    mca = True
                    nmca = end_chan - start_chan + 1
                    mca_data = np.zeros(nmca)
                    mca_i = 0
                
                if start and mca:
                    if mca_i == 0:
                        ifst = 3
                    else:
                        ifst = 1
                    
                    if line.endswith('\\'):
                        line = line[:-1]
                    
                    tdata = list(map(int, line[ifst:].split()))
                    mca_data[mca_i:mca_i + len(tdata)] = tdata
                    mca_i += len(tdata)
                    
                    if mca_i == nmca:
                        M_mca_data.append(mca_data.copy())
                        mca = False
                    continue
                
                if start and not mca and not line.startswith('#'):
                    tdata = list(map(float, line.split()))
                    if len(tdata) == len(cnames):
                        data.append(tdata)
                
                if start and line.startswith('#S') and f'#S {scanno} ' not in line:
                    start = False
            
            pdata = pd.DataFrame(data, columns=cnames)
            scan_info = {'columns': cnames[1:], 'title': title}
        
        return pdata, scan_info, M_mca_data, start_chan, end_chan
    
    except FileNotFoundError:
        print(f"Error: The file '{specfile}' was not found.")
        return None, None, None, None, None
    except Exception as e:
        print(f"An error occurred: {e}")
        return None, None, None, None, None

def getscan_mca(filename,scan_no):
    data, scan_info,M_mca_data,start_ch,end_ch = readscan_2(filename,scan_no)
    x = data['Two Theta']
    y = data['Detector']
    mon = data['Seconds']
    dy = np.sqrt(y)
    y /= mon
    dy /= mon
    y_mca = y*0
    for i,mca_data in enumerate(M_mca_data):
        y_mca[i] = np.sum(mca_data)
    dy_mca = np.sqrt(y_mca)
    y_mca/=mon
    dy_mca /= mon
    return x,y,dy,y_mca,dy_mca

def merge_scans(specfile,scanset,norm='mca',mca=False):
    x = np.array([])
    y = np.array([])
    dy = np.array([])
    for scanno, bg, att in scanset:
        if bg:
            if mca:
                tx,ty,tdy = getscan_bg_mca(specfile,scanno,norm)
            else:
                tx,ty,tdy = getscan_bg(specfile,scanno,norm)
            ty *= att
            tdy *= att
        else:
            if mca:
                tx,ty,tdy,y_mca,dy_mca = getscan_mca(specfile,scanno)
            else:
                tx,ty,tdy = getscan(specfile,scanno, norm)
            ty *= att
            tdy *= att
        x = np.append(x,tx)
        y = np.append(y,ty)
        dy = np.append(dy,tdy)
    ind = np.argsort(x)
    x = x[ind]
    y = y[ind]
    dy = dy[ind]
    return x,y,dy

def getscan_bg_mca(filename,scan_no,norm):
    x,y,dy,x_mca,y_mca = getscan_mca(filename,scan_no)
    x,yb1,dyb1,x_mca,y_mca = getscan_mca(filename,scan_no+1)
    x,yb2,dyb2,x_mca,y_mca = getscan_mca(filename,scan_no+2)
    ys = y-(yb1+yb2)/2
    dys = np.sqrt(dy**2 + dyb1**2 + dyb2**2)
    return x,ys,dys


def merge_duplicates(x,y,dy):
    # sort and reorder x
    w = np.argsort(x)
    x=x[w]
    y=y[w]
    dy=dy[w]
    # now look for duplicates
    wdup = x[0:-1]-x[1:]==0
    while np.sum(wdup)>0:
        ofst = 0
        for j in np.argwhere(wdup):
            i = int(j-ofst)
            y[i],dy[i] = cbwe_s(y[i],dy[i],y[i+1],dy[i+1])
            y = np.append(y[0:i+1],y[i+2:])
            x = np.append(x[0:i+1],x[i+2:])
            dy = np.append(dy[0:i+1],dy[i+2:])
            ofst += 1
        wdup = x[0:-1]-x[1:]==0
    return(x,y,dy)
        
def getscan(filename,scan_no,norm):
    data, scan_info = readscan(filename,scan_no)
    x = data['Two Theta']
    y = data['Detector']
    if norm == 'mca':
        mon = data['mca']
        if np.mean(mon)<1000:
            mon = mon*0+np.mean(mon)
        mon = mon*np.mean(data['Seconds']/data['mca'])
    else:
        mon = data['Seconds']
    dy = np.sqrt(y)
    y /= mon
    dy /= mon
    return x,y,dy

def getscan_bg(filename,scan_no,norm):
    x,y,dy = getscan(filename,scan_no,norm)
    x,yb1,dyb1 = getscan(filename,scan_no+1,norm)
    x,yb2,dyb2 = getscan(filename,scan_no+2,norm)
    ys = y-(yb1+yb2)/2
    dys = np.sqrt(dy**2 + dyb1**2 + dyb2**2)
    return x,ys,dys


def list_scans(filename,start_no=0, end_no=np.Inf):
    try:
        with open(filename,'r') as fd:
            for line in fd:
                lar = line.split()
                if len(lar) > 0:
                    if lar[0] == '#S':
                        scan_no = int(lar[1])
                        if scan_no > start_no and scan_no < end_no:
                            print(f'{line.strip():s}')
    except FileNotFoundError:
        print(f"Error: The file '{filename}' was not found.")
    except IOError:
        print(f"Error: An error occurred while reading the file '{filename}'.")


def get_reflectivity_CLS(fname, datadir, scan):
    specfile = datadir + fname
    if os.path.isfile(specfile):
        data,scan_info  = readscan(specfile,scan)
        y0 = data['L_ROI1']
        bg = (data['L_ROI2']+data['L_ROI3'])/2 
        y = y0-bg
        dy = np.sqrt(y0 + bg)
        x = data['MU']
    else:
        print(f'File {specfile} does not exist')
    return(x, y,dy)

def get_refl_sequence_CLS(fname, datadir, firstscan):
    norm = 2.02e-3
    x1, y1, dy1 = get_reflectivity_CLS(fname,datadir, firstscan)
    y1 /= norm 
    dy1 /= norm
    x2, y2, dy2 = get_reflectivity_CLS(fname,datadir, firstscan +1)
    y = np.concatenate((y1,y2))
    dy = np.concatenate((dy1,dy2))
    x = np.concatenate((x1,x2))
    x, y, dy = merge_duplicates(x,y,dy)
    return x, y, dy

def plot_refl_sequence_CLS(fname, datadir, firstscan, k, plotdip = True):
    x, y, dy = get_refl_sequence_CLS(fname, datadir, firstscan)
    if k:
        q = 2*k*np.sin(x*scc.degree)
        plt.errorbar(q,y,dy,fmt='k.')
        plt.xlabel('q (inv ang)')
        if plotdip:
            arrow = FancyArrowPatch((.07, 1e7), (.07,1e5),
            arrowstyle='-|>',  # Line and arrow head
            mutation_scale=20,  # Size of the arrow head
            color='black', linewidth=1)
            ax = plt.gca()
            ax.add_patch(arrow)
            plt.text(.07,1e7,'expected dip')
    else:
        plt.errorbar(x,y,dy,fmt='k.')
        plt.xlabel('mu (deg)')
    plt.ylabel('Lambda counts')
    plt.yscale('log')
    plt.title(f'{fname:s} scans {firstscan:d} - {firstscan+1:d}')

def half(q,I,dI):
    '''
    and merges all pairs of adjacent points in q, 
    so that there are half as many q points with 
    correspondingly smaller error bars.  
    '''
    nlen = int(2*np.floor(len(q)/2))
    q = q[0:nlen]
    I = I[0:nlen]
    dI = dI[0:nlen]
    eve = np.arange(0,nlen-1,2).astype(int)
    odd = np.arange(1,nlen,2).astype(int)
    qout = (q[eve]+q[odd])/2
    Iout,dIout = cbwe(I[eve],dI[eve],I[odd],dI[odd])
    return  qout,Iout,dIout 

def cbwe(y1,dy1,y2,dy2):
    if (dy1 == 0 ) and (dy2 == 0):
        return (y1+y2)/2, 0
    elif (dy1 == 0):
        return y2, dy2
    elif (dy2 == 0):
        return y1, dy1
    else:
        lg = len(y1)
        y = (y1/dy1**2+y2[0:lg]/dy2[0:lg]**2)/(1/dy1**2+1/dy2[0:lg]**2)
        dy = np.sqrt(1/(1/dy1**2 + 1/dy2[0:lg]**2))
        np.append(y,y2[lg:])
        np.append(dy,dy2[lg:])
        return y, dy

def cbwe_s(y1,dy1,y2,dy2):
    # cbwe for scalers
    y = (y1/dy1**2+y2/dy2**2)/(1/dy1**2+1/dy2**2)
    dy = np.sqrt(1/(1/dy1**2 + 1/dy2**2))
    return y,dy

MCA_DATASIZE = 2048

def get_mca_data_CLS_Nov(mca_filename,scan_number,exposure,quiet=False):
    ''' 
	  Routine to met mca data from November 2023 CLS data
	  February 2024 data changed format so this no longer works
	  quite suppresses printout of dead time percentage
    returns mca data as a two dimensional array of scan point and mca data
    also returns number of points in the dataset (ndset)
    '''
    df = pd.read_csv(f'{mca_filename}_{scan_number}.txt',sep=' ',header=None)
    xy = df.to_numpy() 
    y= np.array(xy[:,1])
    ndset = int(np.size(y)/MCA_DATASIZE)
    mca = np.reshape(y,(ndset,MCA_DATASIZE))
    # correct data for deadtime
    for i in range(ndset):
        mca[i,:] = dtcorrect(mca[i,:],exposure,quiet)
    return(mca,ndset)

def get_DND_refl(data):
    I = data['ic_refl'].to_numpy()
    th = data['Kphi'].to_numpy()
    norm = data['IC5IDB']
    exposure_time = np.mean(data['Seconds'])
    return th, I, exposure_time, norm

def get_CLS_refl(data):
    I = data['L_ROI1'].to_numpy()
    th = data['MU'].to_numpy()
    norm = data['IC2']
    exposure_time = np.mean(data['Seconds'])
    return th, I, exposure_time, norm

def get_fluor_data(dinfo,  scan_number, quiet=True):
    data,scan_info  = readscan(dinfo['spec_filename'],scan_number)
    if dinfo['run_date'] == 'Feb24':
        mca,Energy = get_mca_data_CLS('',dinfo['mca_filename'],scan_number,quiet=quiet)
        angles, I, exposure_time, norm = get_CLS_refl(data)
    elif dinfo['run_date'] == 'Nov23':
        mca,ndset = get_mca_data_CLS_Nov(dinfo['mca_filename'],scan_number,exposure_time,quiet=quiet)
        angles, I, exposure_time, norm = get_CLS_refl(data)
    elif dinfo['run_date'] == 'Mar23':
        mca = get_mca_data_DND(dinfo['mca_filename'],scan_number)
        angles, I, exposure_time, norm = get_DND_refl(data)
    else:
        print(f'Not configured for run_date {dinfo["run_date"]}')
        sys.exit() 
  
    mca_sum = np.sum(mca,0) # sum over each point in the scan
    ndset = np.shape(mca)[0]
    npt = np.shape(mca)[1]
    return mca,mca_sum,ndset,angles,npt,exposure_time,norm

def fit_fluor_data(mca, ndset, npt, exposure_time, norm, peak_model, pars, E, data_range):
    ''' Fit each mca set to a model for peak and return array of amplitudes'''
    amplitude_array = np.zeros(npt)
    amplitude_uncertainty = np.zeros(npt)

    mean_norm = np.mean(norm) / norm  # Compute once and use for all iterations
    X = E[data_range]
    summed_data =  np.sum(mca,0)[data_range]
    global_result =  peak_model.fit(summed_data, x=X, params=pars)
    global_result.plot_fit()
    #print(global_result.fit_report())

    for mca_scan_number in range(ndset):
        single_data_set = mca[mca_scan_number, :][data_range]
        
        this_result = peak_model.fit(single_data_set, x=X, params=pars)

        amplitude_value = this_result.params['amplitude'].value / exposure_time
        amplitude_error = this_result.params['amplitude'].stderr if isinstance(this_result.params['amplitude'].stderr, float) else 0
        amplitude_error /= exposure_time

        amplitude_array[mca_scan_number] = amplitude_value * mean_norm[mca_scan_number]
        amplitude_uncertainty[mca_scan_number] = amplitude_error * mean_norm[mca_scan_number]

    return amplitude_array, amplitude_uncertainty


def get_mca_data_DND(specfile,scanno):
    with open(specfile,'r') as fp:
        # read down to start of scan
        start = False
        mca = False
        mca_i = 0
        nmca = 0
        mca_data = []
        M_mca_data = []
        start_chan = 0
        end_chan = 2047
        channels = 2048
        data = []
        for i, line in enumerate(fp):
            if not start and '#S {0:d} '.format(scanno) in line:
                print(f'found scan {scanno:d} at line {i:d}')
                start = True
                title = line
                continue
            if start and '#L' in line:
                cnames = re.split('\s{2,}',line[3:-1]) 
                continue
            if start and '@0' in line:
                mca = True
                nmca = (end_chan-start_chan+1)
                mca_data = np.zeros(nmca)
                mca_i = 0          
            if start and mca:
                line0=line
                if mca_i == 0:
                    ifst = 1
                else:
                    ifst = 1
                if ord(line[-2]) == 92:
                    line = line[:-2]
                tdata = list(map(int,line[ifst:].split())) 
                mca_data[mca_i:min(mca_i+32,nmca)] = tdata
                mca_i += np.size(tdata)
                if mca_i == nmca:
                    M_mca_data.append(mca_data)
                    mca = False
                continue
            if start  and '#S' in line:
                start = False
    return(np.array(M_mca_data))

def dtcorrect(y,exposure,quiet):
    ''' Correct Dead Time for CLS MCA'''
    N = np.sum(y)/exposure
    tau = 4.1e-6
    N0 = fsolve(lambda N0: N - N0*np.exp(-N0*tau),N)[0] 
    if not quiet:
        print('percent dead time: {0:2.0f}'.format((1-N/N0)*100))
    return y*N0/N

def get_mca_data_CLS(dir,fname,scan_nu,quiet=True):
    # exposure is the exposure time of the mca
    # If not quiet, then dtcorrect prints the dead time
    with open(dir+fname,'r') as fd:
        mca_size = 2048
        for line1 in fd:
            if line1.startswith(f"#S {scan_nu:d}"):
                print(f'found scan {scan_nu:d}')
                line2 = fd.readline()
                line3 = fd.readline()
                line4 = fd.readline()
                line5 = fd.readline()
                line6 = fd.readline()
                npt = int(line1.split()[-2])+1
                mca_size = int(line3.split()[1])
                exposure = float(line4.split()[1])
                pnum = int(line5.split()[1])
                mca = zeros([npt,mca_size])
                Energy = zeros(mca_size)
                for i in range(mca_size):
                    nline = fd.readline()
                    Energy[i] =float(nline.split()[1])
                    mca[0,i] = int(nline.split()[2])
                for j in range(npt-1):
                    for k in range(8):
                        fd.readline()
                    for l in range(mca_size):
                        mca[j+1,l] = int(fd.readline().split()[2])
                break
    for i in range(npt):
                mca[i,:] = dtcorrect(mca[i,:],exposure,quiet)
    return mca,Energy

def plot_mca_sum(datadir,fname,snum,xmin=1400,xmax=14800,scale='log'):
    mca, Energy = get_mca_data_CLS(datadir,fname,snum,True)
    mca_sum = np.sum(mca,0)
    rr = (Energy>xmin)*(Energy<xmax)
    plt.plot(Energy[rr],mca_sum[rr],'-k')
    plt.xlabel('Energy (keV)')
    plt.ylabel('counts')
    plt.yscale(scale)
    plt.title(f'{fname:s} scan {snum:d}')

def peak_label(Energy,Info,height=.8,linespec='-r'):
    ax = plt.gca()
    y_min, y_max = ax.get_ylim()
    
    # Determine if the y-axis is log scale
    is_log = ax.get_yscale() == 'log'
    
    if is_log:
        # Calculate the y-values for 20% and 80% of the y-axis range in log scale
        y_start = np.power(10, np.log10(y_min) + (1-height) * (np.log10(y_max) - np.log10(y_min)))
        y_end = np.power(10, np.log10(y_min) + height * (np.log10(y_max) - np.log10(y_min)))
    else:
        # Calculate the y-values for 20% and 80% of the y-axis range in linear scale
        y_start = y_min + (1-height) * (y_max - y_min)
        y_end = y_min + height * (y_max - y_min)
    
    # Draw the vertical line using ax.plot
    ax.plot([Energy, Energy], [y_start, y_end], fmt )
    ax.text(Energy,y_end,Info)

def peak_label(Energy,Info,height=.8,linespec='-r'):
    ax = plt.gca()
    y_min, y_max = ax.get_ylim()
    
    # Determine if the y-axis is log scale
    is_log = ax.get_yscale() == 'log'
    
    if is_log:
        # Calculate the y-values for 20% and 80% of the y-axis range in log scale
        y_start = np.power(10, np.log10(y_min) + .1 * (np.log10(y_max) - np.log10(y_min)))
        y_end = np.power(10, np.log10(y_min) + height * (np.log10(y_max) - np.log10(y_min)))
    else:
        # Calculate the y-values for 20% and 80% of the y-axis range in linear scale
        y_start = y_min + .1 * (y_max - y_min)
        y_end = y_min + height * (y_max - y_min)
    
    # Draw the vertical line using ax.plot
    ax.plot([Energy, Energy], [y_start, y_end], linespec )
    ax.text(Energy,y_end,Info)

def K_label(elem,height=.8):
    ax = plt.gca()
    x_min, x_max = ax.get_xlim()
    lines = xdb.xray_lines(elem,'K')
    lines = ['Ka1','Kb1']
    nlab = 0
    for line in lines:
        try:
            lE = xdb.xray_lines(elem,'K')[line][0]
            N2 = xdb.xray_lines(elem,'K')[line][1]
            if (lE > x_min) and (lE < x_max):
                if nlab == 0:
                    peak_label(lE,elem,linespec='-r',height=height)
                    nlab += 1
                else:
                    peak_label(lE,'',linespec='--r',height=height)
        except:
            continue

def L_label(elem,height=.8):
    ax = plt.gca()
    x_min, x_max = ax.get_xlim()
    edges = ['L1','L2','L3']
    for edge in edges:
        N1 = xdb.xray_edge(elem,edge)[1]
        N1 *= xdb.xray_edge(elem,edge)[2]-1
        lines = xdb.xray_lines('Au',edge)

        for line in lines:
            try:
                lE = xdb.xray_lines(elem,edge)[line][0]
                N2 = xdb.xray_lines(elem,edge)[line][1]
                if (N2*N1 > .02) and (lE > x_min) and (lE < x_max):
                    peak_label(lE,elem,linespec='-y', height=height)
            except:
                continue

def plot_br_fluor():
    DeltaE = (Energy[-1]-Energy[0])/2048
    dms = np.shape(mcas)
    y = np.zeros(dms[0])
    dy = np.zeros(dms[0])
    par = Br_peak_mod.make_params()
    par['A_Au'].value=1e5
    par['A_Br'].value=1e5
    par['sig'].value=111
    par['sig'].vary = 0
    par['intercept'].value=0
    par['slope'].value=0
    result = Br_peak_mod.fit(np.sum(mcas,0)[rr],x=Energy[rr],params=par)
    #nresult.plot_fit()
    par = result.params;
    par['A_Au'].value /= dms[0]
    par['A_Br'].value /= dms[0]
    for i,mca in enumerate(mcas):
        result = Br_peak_mod.fit(mca[rr],x=Energy[rr],params=par)
        y[i] = result.params['A_Br'].value/DeltaE
        dy[i] = result.params['A_Br'].stderr/DeltaE
    return y, dy

def get_br_amps(Energy,mcas,rr,plot=False):
    DeltaE = (Energy[-1]-Energy[0])/2048
    dms = np.shape(mcas)
    y = np.zeros(dms[0])
    dy = np.zeros(dms[0])
    par = Br_peak_mod.make_params()
    par['A_Au'].value=1e5
    par['A_Br'].value=1e5
    par['sig'].value=111
    par['sig'].vary = 0
    par['intercept'].value=0
    par['slope'].value=0
    result = Br_peak_mod.fit(np.sum(mcas,0)[rr],x=Energy[rr],params=par)
    par = result.params;
    par['A_Au'].value /= dms[0]
    par['A_Br'].value /= dms[0]
    for i,mca in enumerate(mcas):
        result = Br_peak_mod.fit(mca[rr],x=Energy[rr],params=par)
        y[i] = result.params['A_Br'].value/DeltaE
        dy[i] = result.params['A_Br'].stderr/DeltaE
    return y, dy

def plot_br_fluor(datadir,fspec,scans):
    fvort =fspec+"_Vortex.mca"
    data,scan_info  = readscan(datadir+fspec,scans[0])
    mu = data['MU'].to_numpy()
    for i, snum in enumerate(scans):
        if i==0:
            mcas,Energy = get_mca_data_CLS(datadir,fvort,snum,True)
        else:
            tmcas,Energy = get_mca_data_CLS(datadir,fvort,snum,True)
            mcas += tmcas
    rr = (Energy>11000)*(Energy<12500)
    y,dy = get_br_amps(Energy,mcas,rr)
    plt.errorbar(mu,y,dy,fmt='ks')
    plt.xlabel('mu (deg)')
    plt.ylabel('Br fluorescence (fit) ')
    scanlist = ''
    for snum in scans:
        scanlist += f'{snum:d} '
    plt.title(f'{fspec:s} scans {scanlist:s}')
    
def get_edge_absorb(element,edge):
    Ee = xdb.xray_edge(element, edge, energy_only=True)
    delE = 50
    xlow = xdb.incoherent_cross_section_elam(element, Ee-delE)
    xhigh = xdb.incoherent_cross_section_elam(element, Ee+delE)
    del_x = xhigh-xlow 
    return del_x

def Au_L_peak(A,sig0,energy):
    E0 = 14500
    element = 'Au'
    edges = ['L1','L2','L3']
    y = energy*0
    norm = 0
    for edge in edges:
        # First get edge amplitude
        lines = xdb.xray_lines(element,edge)
        N1 = get_edge_absorb(element,edge)
        for line in lines:
                lE = xdb.xray_lines(element,edge)[line][0]
                N2 = xdb.xray_lines(element,edge)[line][1]
                N3 = xdb.fluor_yield(element, edge, line, E0)[0]
                sig = sig0*energy/Br_ka 
                arg = (energy-lE)**2/2/sig**2
                amp = 1/np.sqrt(2*np.pi*sig**2)
                y += N1*N2*N3*amp*np.exp(-arg)
                norm += N1*N2*N3
    y /= norm
    y *= A
    return y
Br_ka = xdb.xray_lines('Br','K')['Ka1'][0]

def Br_K_peak(A,sig,energy):

    arg = (energy-Br_ka)**2/2/sig**2
    amp = A/np.sqrt(2*np.pi*sig**2)
    y = amp*np.exp(-arg)
    return y

def Br_peak_sim(x,A_Br,A_Au,sig):
     y = Br_K_peak(A_Br,sig,x)
     y += Au_L_peak(A_Au,sig,x)
     return y

Br_background = LinearModel()
Br_peak_mod = Model(Br_peak_sim)+Br_background




# -*- coding: utf-8 -*-
"""
multilayer_fluor_fit

file to fit fluorescence from bilayer on top of multilayer

Created on Tue Mar 14 13:50:10 2023
Code to generate multilayer sample simulation


@author: lluri
"""
import scipy.constants as scc
from pySWXF.refl_funs import reflection_matrix, standing_wave 
from pySWXF.xray_utils import eden_to_rho
from lmfit import Model
from scipy.special import erf
import numpy as np
import time
import xraydb as xdb
import scipy.constants as scc
import scipy.signal as scs 
#%% Imports and configuration section
# -*- coding: utf-8 -*-
"""
Created on Tue May  9 12:40:26 2023

@author: th0lxl1
"""
#%%
def make_bilayer_model():
    Head = 'CH2'
    Tail = 'CH2'
    Methyl = 'CH2'
    rho_head = eden_to_rho(Head, 486/scc.nano**3)
    rho_tail = eden_to_rho(Tail, 323/scc.nano**3)
    bilayer_layers = [(Head,rho_head,7,3.17),            # Water                                                                          
              (Tail,rho_tail,17.23,3.17),       # Distal Acyl Chain
              (Methyl,0,2,3.17),                # Methyl Overlap Region
              (Tail,rho_tail,17.63,2.6),         # Proximal Acyl Chain
              (Head,rho_head,7,2.6)]           # Proximal Head Group
    return bilayer_layers
#%%

def sample_sim(G,D,sig_Mo,sig_Si,overlay,
               sig_Si_top,water,bilayer,
               D_Ti=108.2,sig_Ti=3.0,D_SiO2=10,sig_SiO2=3,D_SiO2_Bot=30,
               sig_SiO2_Bot=3,sig_Si_Bot=3, **kwargs):
    if water:
        layers = [('H2O',1.0,0,0)]  
        if bilayer:
            layers += make_bilayer_model()
    else:
        layers = [('N2',1.225e-3,0,0)]
    layers += [('SiO2',eden_to_rho('SiO',692e27),D_SiO2,sig_SiO2),
    ('SiO',eden_to_rho('SiO',560e27),1.5,1.5,),
    ('Si',xdb.atomic_density('Si'),overlay,sig_Si_top)]
    nstack = 20
    rMo = xdb.atomic_density('Mo')
    rSi = xdb.atomic_density('Si')
    for i in range(nstack):
        layers += [('Mo',rMo,D*G,sig_Mo),
                       ('Si',rSi,D*(1-G),sig_Si)]
    layers += [('Ti',xdb.atomic_density('Ti'),D_Ti,sig_Ti), 
                   ('SiO2',2.2,D_SiO2_Bot,sig_SiO2_Bot),
                   ('Si',xdb.atomic_density('Si'),0,sig_Si_Bot)]
    return layers

def multilayer_ref_Ti_orig(theta,I0,thoff,
                   bg,foot,G,D,sig_Mo,sig_Si,overlay,
                   sig_Si_top,water,
                   Energy,bilayer,res,D_Ti,sig_Ti,D_SiO2,sig_SiO2,
                   D_SiO2_Bot,sig_SiO2_Bot,sig_Si_Bot):
    if res == 0:
        norm = 1
        nspan = 1
        dth_span = [0]
        mul = [1]
        # print('multilayer_ref_Ti: bypassing resolution')
    else:
        nspan = 8
        dth_span = np.linspace(-2*res,2*res,nspan)
        mul = np.exp(-dth_span**2/2/res**2)
        norm = np.sum(mul)
        # print('resolution = {0:7.2f} nspan = {1:d}'.format(res,nspan))
    vals = locals()
    tic = time.time()
    result_list = []
    for i in range(nspan):
        dth = vals['dth_span'][i]
        alpha = (vals['theta'] - vals['thoff'] + dth)*scc.degree
        layers = sample_sim(G,D,sig_Mo,
                sig_Si,overlay,
                sig_Si_top,water,bilayer,
                D_Ti = D_Ti, sig_Ti = sig_Ti,
                D_SiO2 = D_SiO2,sig_SiO2=sig_SiO2,
                D_SiO2_Bot = D_SiO2_Bot, sig_SiO2_Bot = sig_SiO2_Bot,
                sig_Si_Bot = sig_Si_Bot)
        t,r,kz,_,_,_,zm = reflection_matrix(
            alpha,Energy,layers)
        Intensity = np.abs(r[:,0])**2
        frange = theta<foot
        Intensity[frange] *= theta[frange]/foot
        tres = Intensity*mul[i]/norm
        result_list.append(tres)       
    toc = time.time()
    #print(f'{toc-tic:2.1f} ',end='')
    return(I0*np.sum(np.array(result_list),0)+bg)

import numpy as np
import scipy.constants as scc
import scipy.signal as scs

def multilayer_ref_Ti(
    theta: np.ndarray,
    I0: float,
    thoff: float,
    bg: float,
    foot: float,
    G: float,
    D: float,
    sig_Mo: float,
    sig_Si: float,
    overlay: float,
    sig_Si_top: float,
    water: bool,
    Energy: float,
    bilayer: bool,
    res: float,
    D_Ti: float,
    sig_Ti: float,
    D_SiO2: float,
    sig_SiO2: float,
    D_SiO2_Bot: float,
    sig_SiO2_Bot: float,
    sig_Si_Bot: float
) -> np.ndarray:
    theta = np.array(theta)
    theta_off = theta - thoff
    
    # Recast data on uniform grid
    dth = np.min(np.diff(theta_off))
    npt = int((theta_off[-1] - theta_off[0]) / dth)
    thnew = np.linspace(theta_off[0], theta_off[-1], npt)
    
    # Simulate sample layers
    layers = sample_sim(G, D, sig_Mo, sig_Si, overlay, sig_Si_top, water, bilayer,
                        D_Ti=D_Ti, sig_Ti=sig_Ti, D_SiO2=D_SiO2, sig_SiO2=sig_SiO2,
                        D_SiO2_Bot=D_SiO2_Bot, sig_SiO2_Bot=sig_SiO2_Bot, sig_Si_Bot=sig_Si_Bot)
    alpha = thnew * scc.degree
    
    # Compute reflection matrix
    _, r, _, _, _, _, _ = reflection_matrix(alpha, Energy, layers)
    Intensity = np.abs(r[:, 0])**2
    
    # Correct for footprint
    frange = thnew < foot
    Intensity[frange] *= thnew[frange] / foot
    
    # Include resolution using convolution
    if res > 0:
        signal = np.concatenate([Intensity[::-1], Intensity, Intensity[::-1]])
        sig = res / dth
        nwin = 8 * int(sig) + 16
        win = np.exp(-0.5 * (np.arange(-nwin // 2, nwin // 2 + 1) / sig)**2)
        win /= win.sum()
        Intensity = scs.convolve(signal, win, mode='same')[npt:2*npt]
    
    # Re-interpolate data back to old angles
    Iout = np.interp(theta_off, thnew, Intensity)
    return I0 * Iout + bg


def multilayer_fluor(theta,I0,thoff,
                   bg,foot,G,D,sig_Mo,sig_Si,overlay,
                   sig_Si_top,sep,hoff,res,
                   Energy):
    water = True
    bilayer = True 
    D_bilayer = 50.86
    if res == 0:
        norm = 1
        nspan = 1
        dth_span = [0]
        mul = [1]
    else:
        nspan = 4
        dth_span = np.linspace(-1.5*res,1.5*res,nspan)
        mul = np.exp(-dth_span**2/2/res**2)
        norm = np.sum(mul)
    for i in range(nspan): 
        dth = dth_span[i]
        alpha = (theta-thoff+dth)*scc.degree 
        layers = sample_sim(G,D,sig_Mo,sig_Si,overlay,
                sig_Si_top,water,bilayer)
        t,r,kz,_,_,_,zm = reflection_matrix(alpha,Energy,layers)
        heights = np.array([sep/2,-sep/2])+hoff - D_bilayer/2
        I,E = standing_wave(heights*scc.angstrom,t,r,kz,zm)
        I = np.average(I,0)
        I /= (theta-thoff)
        I *= np.mean(theta)*I0
        if i==0:
            Itot = I*mul[i]/norm
        else:
            Itot += I*mul[i]/norm 
    Itot[(theta-thoff)<foot] *= (theta-thoff)[(theta-thoff)<foot]/foot
    Itot += bg
    print('.',end='')
    return Itot


multilayer_model_Ti = Model(multilayer_ref_Ti, independent_vars=['theta','Energy','water','bilayer'])
fluor_model = Model(multilayer_fluor, independent_vars=['theta','Energy'])

xdb.add_material('silicon_carbide', 'Si1C1', 3.21, categories=None)
SiC = xdb.find_material('silicon_carbide')
W = xdb.find_material('W')
Cr = xdb.find_material('Cr')
Si = xdb.find_material('Si')
H2O = xdb.find_material('water')
air = xdb.find_material('air')
xdb.add_material('silicon_carbide', 'Si1C1', 3.21, categories=None)
kapton = xdb.find_material('kapton')


def sample_sim_new(d_overlay, sig_overlay, G,D,sig_W,sig_SiC,
                   D_Cr, sig_Cr, sig_Si_Bot, water=False, bilayer = True, new_cell = True):
    if new_cell:
        layers = [(air.formula,air.density,0,0)]
        d_kapton = 1000 # not correct, but big
        d_water = 1000
        sig_kapton = 10 # just a random guess
        layers += [(kapton.formula, kapton.density,d_kapton, sig_kapton)]
        layers += [(H2O.formula, H2O.density,d_water, sig_kapton)]
        if bilayer:
            layers += make_bilayer_model()
    elif water:
        layers = [(H2O.formula,H2O.density,0,0)]
        if bilayer:
            layers += make_bilayer_model()
    else:
        layers = [(air.formula,air.density,0,0)]

    # now put in cap layer
    layers += [(SiC.formula,SiC.density,d_overlay,sig_overlay)]
    # put in multilayer stack
    nstack = 15
    for i in range(nstack):
        layers += [(W.formula,W.density,D*G,sig_W)]
        layers += [(SiC.formula,SiC.density,D*(1-G),sig_SiC)]
    # add chrome binding layer
    layers += [(Cr.formula,Cr.density,D_Cr,sig_Cr)]
    # add substrate
    layers += [(Si.formula,Si.density,0,sig_Si_Bot)]
    return layers



def multilayer_ref_new_orig(theta,I0,thoff,bg,foot,G,D,d_overlay, sig_overlay,
    sig_W,sig_SiC,Energy,res,D_Cr,sig_Cr,sig_Si_Bot, water=True,bilayer=True, new_cell=True):
    theta = np.array(theta)
    theta_off = theta - thoff
    #recast data on uniform grid
    dth = min(theta_off[1:]-theta_off[0:-1])
    npt = int((theta_off[-1]-theta_off[0])/dth)
    thnew = np.linspace(theta_off[0],theta_off[-1],npt)
    layers = sample_sim_new(d_overlay, sig_overlay, G,D,sig_W,sig_SiC,
                   D_Cr, sig_Cr, sig_Si_Bot,water=water,bilayer=bilayer,new_cell=new_cell)
    alpha = thnew*scc.degree
    t,r,kz,_,_,_,zm = reflection_matrix(alpha,Energy,layers)
    Intensity = np.abs(r[:,0])**2
    # Correct for footprint
    frange = thnew<foot
    Intensity[frange] *= thnew[frange]/foot
    # Include resolution
    if (res>0):
        npt = np.size(Intensity)
        signal = np.concatenate((Intensity[::-1],Intensity,Intensity[::-1]))
        step = dth
        sig = res/step
        nwin = 8*int(sig)+16
        x2 = np.arange(-nwin/2,nwin/2+1)
        win = np.exp(-x2**2/2/sig**2)
        win /= np.sum(win)
        signal2 = scs.convolve(signal,win,mode='same')
        signal2 = signal2[npt:2*npt]
        Intensity = signal2
    # re-interpolate data back to old angles
    Iout = np.interp(theta_off,thnew,Intensity)
    return(I0*Iout+bg)

import numpy as np
import scipy.constants as scc
import scipy.signal as scs



def multilayer_ref_new(
    theta: np.ndarray,
    I0: float,
    thoff: float,
    bg: float,
    foot: float,
    G: float,
    D: float,
    d_overlay: float,
    sig_overlay: float,
    sig_W: float,
    sig_SiC: float,
    Energy: float,
    res: float,
    D_Cr: float,
    sig_Cr: float,
    sig_Si_Bot: float,
    water: bool = True,
    bilayer: bool = True,
    new_cell: bool = True
) -> np.ndarray:
    theta = np.array(theta)
    theta_off = theta - thoff
    
    # Recast data on uniform grid
    dth = np.min(np.diff(theta_off))
    npt = int((theta_off[-1] - theta_off[0]) / dth)
    thnew = np.linspace(theta_off[0], theta_off[-1], npt)
    
    # Simulate sample layers
    layers = sample_sim_new(d_overlay, sig_overlay, G, D, sig_W, sig_SiC, D_Cr, sig_Cr, sig_Si_Bot, water=water, bilayer=bilayer, new_cell=new_cell)
    alpha = thnew * scc.degree
    
    # Compute reflection matrix
    _, r, _, _, _, _, _ = reflection_matrix(alpha, Energy, layers)
    Intensity = np.abs(r[:, 0])**2
    
    # Correct for footprint
    frange = thnew < foot
    Intensity[frange] *= thnew[frange] / foot
    
    # Include resolution
    if res > 0:
        signal = np.concatenate([Intensity[::-1], Intensity, Intensity[::-1]])
        sig = res / dth
        nwin = 8 * int(sig) + 16
        win = np.exp(-0.5 * (np.arange(-nwin // 2, nwin // 2 + 1) / sig)**2)
        win /= win.sum()
        Intensity = scs.convolve(signal, win, mode='same')[npt:2*npt]
    
    # Re-interpolate data back to old angles
    Iout = np.interp(theta_off, thnew, Intensity)
    return I0 * Iout + bg


multilayer_ref_new_model = Model(multilayer_ref_new,
    independent_vars = ['theta','Energy','water','bilayer','new_cell'])
  

    


# Model for gold fluorescence
from lmfit import Model, models
from matplotlib import pyplot as plt
import numpy as np
import scipy.constants as scc
import importlib.resources
from lmfit import Parameters
import xraydb as xdb
from pySWX.fluor_fit import multilayer_ref_new_model

g1 = models.GaussianModel(prefix='g1')
g2 = models.GaussianModel(prefix='g2')
q1 = models.QuadraticModel() 
peak_model = g1+g2+q1

# code to fit reflectivity and find offset

MULTILAYER_FIT = 'NM_clean_NIU_NOKW_Mar26_24_fit4.json'
def NN_offset_fit(th,I,dI,Beam_Energy, showfit = True):
    with importlib.resources.open_text('pySWXF.data', MULTILAYER_FIT) as f:
    params = Parameters.load(f)

    for key in params.keys():
        params[key].vary = False
    params['I0'].vary = True
    params['I0'].max = 1e9
    params['I0'].min = 100
    params['thoff'].vary = True
    params['thoff'].min =-.1
    params['thoff'].max = .1
    params['res'].value = .001

    alphac_water = np.sqrt(2*xdb.xray_delta_beta('H2O',1.00,Beam_Energy)[0])
    thc_water = alphac_water/scc.degree
    th_refrac = np.sqrt(th**2  - thc_water**2)

    result = multilayer_ref_new_model.fit(
        I,theta = th_refrac, params = params, Energy = Beam_Energy,
        water = True, bilayer = True, new_cell = False, weights = 1/dI)
    
    thoff = result.params['thoff'].value
    
    if showfit:    
        ysim=result.eval(theta = th_refrac, 
            Energy = Energy_Nov, water = True, bilayer = True, new_cell = False)
        result = fluor_fit.multilayer_ref_new_model.fit(
        I,theta = th_refrac, params = params, Energy = Energy_Nov,water = True, bilayer = True, new_cell = False, weights = I*0+1)
        ysim=result.eval(theta = th_refrac, Energy = Energy_Nov, water = True, bilayer = True, new_cell = False)
        plt.plot(th,I,label='data')
        plt.plot(th,ysim,'-k',label='fit')
        plt.locator_params(axis='x', nbins=20)
        plt.grid()
        plt.xlabel('th (deg)')
        plt.ylabel('Reflected Intensity')
    
        plt.legend()
        print(f'angular offset = {thoff:7.3f}')
    return(thoff)
    


def get_gold_amplitude(E,mca_data):   
    pars = peak_model.make_params()
    pars['g1center'].value = 13400
    pars['g1center'].vary = 0
    pars['g2center'].value = 13771
    pars['g2center'].vary = 0
    pars['g1sigma'].value = 114.8
    pars['g1sigma'].vary = 0
    pars['g2sigma'].value = 114.9
    pars['g2sigma'].vary = 0
    pars['g1amplitude'].value = 1e5
    pars['g2amplitude'].value = 2e4
    pars['a'].value = 0
    pars['b'].value = -7.5
    pars['c'].value = 50000
    Erange = (E>13000)*(E<14100)
    result = peak_model.fit(mca_data[Erange],params=pars,x=E[Erange])
    fitpars = result.params
    MCA_SLOPE = E[1]-E[0]
    peak_counts = (fitpars['g1amplitude'].value+fitpars['g1amplitude'].value)/MCA_SLOPE
    peak_errs = np.sqrt((fitpars['g1amplitude'].stderr**2+fitpars['g1amplitude'].stderr**2))/MCA_SLOPE
    return(peak_counts,peak_errs)

def get_Zlist(N,D):
    # D = bilayer thickness
    # N slabs in bilayer
    edgelist = np.linspace(0,D,N+1)        # positions of interfaces of slabs
    Zlist = (edgelist[0:-1]+edgelist[1:])/2   # positions of centers of slabs
    return Zlist, edgelist

def multilayer_fluor_lay_N(theta,Avec,Imap,zmax):
    ''' multilayer_fluor_lay_N(theta,I0,thoff,bg,Avec)
    breaks up bilayer into N slabs wit N the dimension of Avec
    The A's are the amplitudes of the slabs
    '''
    # need to add feature to convolute with angular resolution
    alpha = theta*scc.degree
    Zlist, edgelist = get_Zlist(np.size(Avec), zmax)
    Ifield = Imap(Zlist, alpha)
    # sum up the product of the fluoresence from each slab times the amplitude in the slab
    y = np.sum(Ifield*np.expand_dims(Avec,1),0)
    return(y)

def plot_N_slab_result(result,NUM_SLABS, zmax):
    """
    Plot the fluorophore concentration across three slabs up to a maximum height.

    Parameters:
    result : object containing simulation parameters and results
    zmax : float, the maximum height to consider for plotting
    """
    # Constants
    ANGSTROM = scc.angstrom  # This assumes scc has been properly imported

    # Unpacking parameters
    A = [result.params[f'A{i}'].value for i in range(NUM_SLABS)]
    _, edgelist = get_Zlist(NUM_SLABS, zmax)

    # Check that edgelist is sufficiently long
    if len(edgelist) < NUM_SLABS + 1:
        raise ValueError("edgelist does not contain enough entries.")

    # Plotting
    for i, tA in enumerate(A):
        edge1 = edgelist[i] / ANGSTROM
        edge2 = edgelist[i + 1] / ANGSTROM
        plt.plot([edge1, edge1], [0, tA], '-k')
        plt.plot([edge1, edge2], [tA, tA], '-k')
        plt.plot([edge2, edge2], [tA, 0], '-k')

    plt.xlabel('height ($\\mathrm{\\AA}$)')
    plt.ylabel('fluorophore concentration')
    plt.title('Fluorophore Concentration Profile')

# Model for three slabs

def three_slab(theta,A0,A1,A2,Imap,zmax):
    return multilayer_fluor_lay_N(theta,[A0,A1,A2],Imap,zmax)

three_slab_model = Model(three_slab, independent_vars = ['theta', 'Imap', 'zmax'])

# Model for five slabs
def five_slab(theta,A0,A1,A2,A3, A4, Imap,zmax):
    return multilayer_fluor_lay_N(theta,[A0,A1,A2,A3, A4],Imap,zmax)

five_slab_model = Model(five_slab, independent_vars = ['theta', 'Imap', 'zmax'])